// Parsing

use std::collections::HashMap;

#[derive(Debug, PartialEq)]
pub struct Page {
    id: String,
    body: String,
    footer: Footer,
}

#[derive(Debug, PartialEq)]
pub enum Footer {
    Ending,
    Goto(String),
    Choices(HashMap<u32, Choice>),
}

#[derive(Debug, PartialEq)]
pub struct Choice {
    option: u32,
    flavor: String,
    redirect: String,
    stat_check: Option<StatCheck>,
    stat_change: Option<StatChange>
}

#[derive(Debug, PartialEq)]
pub struct StatCheck {
    stat: String,
    value: i32,
    rel: char,
}

#[derive(Debug, PartialEq)]
pub struct StatChange {
    stat: String,
    addend: i32,
}


peg::parser! {
    grammar storybook_parser() for str {
        pub rule story() -> Vec<Page>
            = blankline()* ps:page()+ eof() { ps }
        rule page() -> Page
            = h:header() b:body() f:footer() blankline()* {
                Page {
                    id: h,
                    body: b,
                    footer: f
                }
            }
        rule header() -> String
            = i:id() __ blankline()* { i }
        rule id() -> String
            = i:$(['A'..='Z'|'_']+) _ { String::from(i) }
        rule body() -> String
            = b:$((!footer() [_])+) { String::from(b.trim()) }
        rule footer() -> Footer
            = "THE END" __ { Footer::Ending } /
              "GOTO" _ i:id() __ { Footer::Goto(i) } /
              c:choice()+ { Footer::Choices(c.into_iter().map(|x| (x.option, x)).collect()) }
        rule choice() -> Choice
            = o:$(['0'..='9']+) ")" _ f:$((!redirect() [_])+) r:redirect() __ {
                let (redirect, stat_change, stat_check) = r;
                Choice {
                    option: o.parse().unwrap(),
                    flavor: String::from(f.trim()),
                    redirect,
                    stat_change,
                    stat_check
                }
            }
        rule redirect() -> (String, Option<StatChange>, Option<StatCheck>)
            = ck:stat_check()? "[" i:id() "]" _ cg:stat_change()? { (i, cg, ck) }
        rule stat_change() -> StatChange
            = "(" a:$(['+'|'-'] ['0'..='9']+) _ n:stat_name() ")" _ { StatChange {
                addend: a.parse().unwrap(),
                stat: n
            }}
        rule stat_check() -> StatCheck
            = "<" n:stat_name() _ v:$(['0'..='9']+) r:['+'|'-'] ">" _ { StatCheck {
                stat: n,
                value: v.parse().unwrap(),
                rel: r
            }}
        rule stat_name() -> String
            = n:$(['a'..='z'|'A'..='Z']+) { String::from(n) }
        rule blankline()
            = _ __
        rule _
            = [' '|'\t']*
        rule __
            = "\r\n" / "\r" / "\n"
        rule eof()
            = ![_]
    }
}

pub struct Book {
    pages: HashMap<String, Page>,
    current_page: Option<String>
}

impl Book {
    pub fn find (&mut self, id: &str) {
        if self.pages.contains_key(id) {
            self.current_page = Some(String::from(id));
        } else {
            self.current_page = None
        }
    }
    pub fn is_ended(&self) -> bool {
        self.current_page == None
    }
    pub fn get_current(&self) -> &Page {
        match &self.current_page {
            None => panic!("Asked for current page, but current page is None!"),
            Some(s) => self.pages.get(s).unwrap(),
        }
    }
    pub fn advance_nooption(&mut self) {
        let footer = &self.get_current().footer;
        match footer {
            Footer::Ending => {self.current_page = None},
            Footer::Goto(s) => {
                let s = s.clone();
                self.find(&s);
            },
            Footer::Choices(_) => panic!("No option for advance on choices page!"),
        }
    }
    pub fn advance_option(&mut self, option: u32) {
        let footer = &self.get_current().footer;
        match footer {
            Footer::Ending => panic!("Option provided for advance on ending page!"),
            Footer::Goto(_) => panic!("Option provided for advance on goto page!"),
            Footer::Choices(choices) => {
                // TODO ban choices with failing stat checks
                // TODO update stats based on stat changes
                match choices.get(&option) {
                    Some(choice) => {
                        let s = choice.redirect.clone();
                        self.find(&s);
                    }
                    None => panic!("Option not a valid choice for page!"),
                }
            }
        }
    }
}

impl From<Vec<Page>> for Book {
    fn from(v: Vec<Page>) -> Book {
        let it = v.into_iter().map(|p| (p.id.clone(), p));
        let pages = HashMap::from_iter(it);
        let mut book = Book {
            pages,
            current_page: None,
        };

        book.find("START");

        book
    }
}

// C API

use std::ffi::CStr;
use std::ffi::CString;
use std::os::raw::c_char;

#[no_mangle]
pub extern fn storybook_make_book(string: *const c_char) -> *mut Book {
    let string = unsafe {
        CStr::from_ptr(string)
    };
    let string = string.to_str().expect("Error, input string not valid UTF-8");
    let pages = storybook_parser::story(string).expect("Error, input string could not be parsed");
    let book = Book::from(pages);
    let book = Box::from(book);
    Box::into_raw(book)
}

#[no_mangle]
pub extern fn storybook_free_book(book: *mut Book) {
    unsafe {
        Box::from_raw(book);
    }
}

#[no_mangle]
pub extern fn storybook_free_string(s: *mut c_char) {
    unsafe {
        drop(CString::from_raw(s));
    }
}

fn get_book_from_ref<'a> (book: *mut Book) -> &'a mut Book {
    let book = unsafe {
        book.as_mut().expect("Invalid Book pointer")
    };
    book
}

#[no_mangle]
pub extern fn storybook_get_body(book: *mut Book) -> *mut c_char {
    let book = get_book_from_ref(book);
    let body = book.get_current().body.clone();
    let body = CString::new(body).expect("Body cannot be converted to CString");
    body.into_raw()
}

#[repr(C)]
pub enum FooterType {
    Ending,
    Goto,
    Choices,
}

#[no_mangle]
pub extern fn storybook_get_footer(book: *mut Book) -> FooterType {
    let book = get_book_from_ref(book);
    match book.get_current().footer {
        Footer::Ending => FooterType::Ending,
        Footer::Goto(_) => FooterType::Goto,
        Footer::Choices(_) => FooterType::Choices,
    }
}

#[no_mangle]
pub extern fn storybook_is_ended(book: *mut Book) -> bool {
    let book = get_book_from_ref(book);
    book.is_ended()
}

#[no_mangle]
pub extern fn storybook_advance_nooption(book: *mut Book) {
    let book = get_book_from_ref(book);
    book.advance_nooption();
    
}
#[no_mangle]
pub extern fn storybook_advance_option(book: *mut Book, option: u32) {
    let book = get_book_from_ref(book);
    book.advance_option(option);
}

#[cfg(test)]
mod tests {
    use super::*;
    const STORY: &str = r"
START

One day, Dane took a really big shit.

GOTO DIALOG

DIALOG

Dane: 'Wow, that was a really big shit!'

1) Die [DIE]
2) Don't die <Strength 100+> [DONTDIE] (-1 Honor)

DIE

Then he died.

Fuck you, Dane!

THE END
";
    #[test]
    fn storybook_parser_test() {
        assert_eq!(
            storybook_parser::story(STORY),
            Ok(vec![
                Page {
                    id: String::from("START"),
                    body: String::from("One day, Dane took a really big shit."),
                    footer: Footer::Goto(String::from("DIALOG"))
                },
                Page {
                    id: String::from("DIALOG"),
                    body: String::from("Dane: 'Wow, that was a really big shit!'"),
                    footer: Footer::Choices(HashMap::from([
                        (1, Choice {
                            option: 1,
                            flavor: String::from("Die"),
                            redirect: String::from("DIE"),
                            stat_check: None,
                            stat_change: None,
                        }),
                        (2, Choice {
                            option: 2,
                            flavor: String::from("Don't die"),
                            redirect: String::from("DONTDIE"),
                            stat_change: Some(StatChange {
                                stat: String::from("Honor"),
                                addend: -1,
                            }),
                            stat_check: Some(StatCheck {
                                stat: String::from("Strength"),
                                value: 100,
                                rel: '+',
                            })
                        }),
                    ]))
                },
                Page {
                    id: String::from("DIE"),
                    body: String::from("Then he died.\n\nFuck you, Dane!"),
                    footer: Footer::Ending,
                }
            ])
        );
    }

    #[test]
    fn test_book_from_pages() {
        let pages = storybook_parser::story(STORY).unwrap();
        let book = Book::from(pages);
        assert_eq!(book.current_page, Some(String::from("START")));
    }
}
